import random
import string

if False:
    from typing import Optional  # noqa: F401

    from apertium_apy import missingdb  # noqa: F401

missing_freqs_db = None  # type: Optional[missingdb.MissingDb]  # has to be global for sig_handler :-/

RECAPTCHA_VERIFICATION_URL = 'https://www.google.com/recaptcha/api/siteverify'
BYPASS_TOKEN = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(24))



from tornado import gen

from apertium_apy.handlers.translate import TranslateHandler
from apertium_apy.utils import to_alpha3_code
from apertium_apy.utils.translation import coreduce


class TranslateChainHandler(TranslateHandler):
    def pair_list(self, langs):
        return [(langs[i], langs[i + 1]) for i in range(0, len(langs) - 1)]

    def get_pairs_or_error(self, langpairs, text_length):
        langs = [to_alpha3_code(lang) for lang in langpairs.split('|')]
        if len(langs) < 2:
            self.send_error(400, explanation='Need at least two languages, use e.g. eng|spa')
            self.log_after_translation(self.log_before_translation(), text_length)
            return None
        if len(langs) == 2:
            if langs[0] == langs[1]:
                self.send_error(400, explanation='Need at least two languages, use e.g. eng|spa')
                self.log_after_translation(self.log_before_translation(), text_length)
                return None
            return self.paths.get(langs[0], {}).get(langs[1])
        for lang1, lang2 in self.pair_list(langs):
            if '{:s}-{:s}'.format(lang1, lang2) not in self.pairs:
                self.send_error(400, explanation='Pair {:s}-{:s} is not installed'.format(lang1, lang2))
                self.log_after_translation(self.log_before_translation(), text_length)
                return None
        return langs

    @gen.coroutine
    def translate_and_respond(self, pairs, pipelines, to_translate, mark_unknown, nosplit=False, deformat=True, reformat=True):
        mark_unknown = mark_unknown in ['yes', 'true', '1']
        chain, pairs = pairs, self.pair_list(pairs)
        for pair in pairs:
            self.note_pair_usage(pair)
        before = self.log_before_translation()
        translated = yield coreduce(to_translate, [p.translate for p in pipelines], nosplit, deformat, reformat)
        self.log_after_translation(before, len(to_translate))
        self.send_response({
            'responseData': {
                'translatedText': self.maybe_strip_marks(mark_unknown, (pairs[0][0], pairs[-1][1]), translated),
                'translationChain': chain,
            },
            'responseDetails': None,
            'responseStatus': 200,
        })
        self.clean_pairs()

    def prepare(self):
        if not self.pairs_graph:
            self.init_pairs_graph()

    @gen.coroutine
    def get(self):
        q = self.get_argument('q', default=None)
        langpairs = self.get_argument('langpairs')
        pairs = self.get_pairs_or_error(langpairs, len(q or []))
        if pairs:
            if not q:
                self.send_response({
                    'responseData': {
                        'translationChain': self.get_pairs_or_error(self.get_argument('langpairs'), 0),
                    },
                    'responseDetails': None,
                    'responseStatus': 200,
                })
            else:
                pipelines = [self.get_pipeline(pair) for pair in self.pair_list(pairs)]
                deformat, reformat = self.get_format()
                yield self.translate_and_respond(pairs, pipelines, q,
                                                 self.get_argument('markUnknown', default='yes'),
                                                 nosplit=False, deformat=deformat, reformat=reformat)
        else:
            self.send_error(400, explanation='No path found for {:s}-{:s}'.format(*langpairs.split('|')))
            self.log_after_translation(self.log_before_translation(), 0)



import re
import os
import logging
try:
    from lxml import etree
except ImportError:
    etree = None

from apertium_apy.utils import to_alpha3_code

if False:
    from typing import Dict, List, Tuple  # noqa: F401


def is_loop(dirpath, rootpath, real_root=None):
    if os.path.islink(dirpath):
        # We just descended into a directory via a symbolic link
        # Check if we're referring to a directory that is
        # a parent of our nominal directory
        if not real_root:
            real_root = os.path.abspath(os.path.realpath(rootpath))
        relative = os.path.relpath(dirpath, rootpath)
        nominal_path = os.path.join(real_root, relative)
        real_path = os.path.abspath(os.path.realpath(dirpath))
        for nominal, real in zip(nominal_path.split(os.sep),
                                 real_path.split(os.sep)):
            if nominal != real:
                return False
        else:
            return True
    else:
        return False


def search_path(rootpath, include_pairs=True, verbosity=1):
    lang_code = r'[a-z]{2,3}(?:_[A-Za-z0-9]+)*'
    type_re = {
        'pair': re.compile(r'({0})-({0})\.mode'.format(lang_code)),
        'analyzer': re.compile(r'(({0}(-{0})?)-(an)?mor(ph)?)\.mode'.format(lang_code)),
        'generator': re.compile(r'(({0}(-{0})?)-gener[A-z]*)\.mode'.format(lang_code)),
        'tagger': re.compile(r'(({0}(-{0})?)-tagger)\.mode'.format(lang_code)),
        'spell': re.compile(r'(({0}(-{0})?)-spell)\.mode'.format(lang_code)),
        'tokenise': re.compile(r'(({0}(-{0})?)-tokenise)\.mode'.format(lang_code)),
        'guesser': re.compile(r'(({0}(-{0})?)-guess(er)?)\.mode'.format(lang_code)),
    }
    modes = {
        'pair': [],
        'analyzer': [],
        'generator': [],
        'tagger': [],
        'spell': [],
        'tokenise': [],
        'guesser': [],
    }  # type: Dict[str, List[Tuple[str, str, str]]]

    real_root = os.path.abspath(os.path.realpath(rootpath))

    for dirpath, dirnames, files in os.walk(rootpath, followlinks=True):
        if is_loop(dirpath, rootpath, real_root):
            dirnames[:] = []
            continue
        for filename in [f for f in files if f.endswith('.mode')]:
            for mtype, regex in type_re.items():
                m = regex.match(filename)
                if m:
                    if mtype != 'pair':
                        modename = m.group(1)  # e.g. en-es-anmorph
                        langlist = [to_alpha3_code(x) for x in m.group(2).split('-')]
                        lang_pair = '-'.join(langlist)  # e.g. en-es
                        dir_of_modes = os.path.dirname(dirpath)
                        mode = (dir_of_modes,
                                modename,
                                lang_pair)
                        modes[mtype].append(mode)
                    elif include_pairs:
                        lang_src = m.group(1)
                        lang_trg = m.group(2)
                        mode = (os.path.join(dirpath, filename),
                                to_alpha3_code(lang_src),
                                to_alpha3_code(lang_trg))
                        modes[mtype].append(mode)

    if verbosity > 1:
        _log_modes(modes)

    return modes


def search_prefs(rootpath):
    if etree is None:
        logging.warning('Please install python3-lxml to enable /pairprefs endpoint')
        return
    real_root = os.path.abspath(os.path.realpath(rootpath))
    prefspath = real_root + '/prefs'
    pairprefs = {}        # type: Dict[str, Dict[str, Dict[str, str]]]
    if not os.path.exists(prefspath):
        return pairprefs
    for f in os.listdir(prefspath):
        fp = os.path.join(prefspath, f)
        try:
            mode = re.sub(r'[.]xml$', '', f)
            pairprefs[mode] = {pref.get('id'): {dsc.get('lang'): dsc.text
                               for dsc in pref.xpath('./description')}
                               for pref
                               in etree.parse(fp).xpath('//preference')}
        except Exception:
            logging.warning('Exception on parsing preferences file {}'.format(fp))
    return pairprefs


def _log_modes(modes):
    """Print given modes to log."""
    for mtype in modes:
        if modes[mtype]:
            logging.info('"%s" modes found:\n%s', mtype, '\n'.join(['\t'.join(m) for m in modes[mtype]]))



import logging
import os
import sqlite3
from concurrent.futures import ThreadPoolExecutor

from tornado import gen

from apertium_apy.handlers.base import BaseHandler
from apertium_apy.utils import to_alpha2_code, iso639_codes  # noqa: F401

if False:
    from typing import Optional  # noqa: F401


# single-threaded for thread-safety
lang_names_db_thread = ThreadPoolExecutor(1)
lang_names_db_conn = None  # type: Optional[sqlite3.Connection]


def get_language_names(locale, db_path):
    global lang_names_db_conn

    if not lang_names_db_conn:
        if os.path.exists(db_path):
            lang_names_db_conn = sqlite3.connect(db_path)
        else:
            return None

    cursor = lang_names_db_conn.cursor()
    return cursor.execute('SELECT * FROM languageNames WHERE lg=?', (locale, )).fetchall()


@gen.coroutine
def get_localized_languages(locale, db_path, languages=[]):
    locale = to_alpha2_code(locale)
    languages = list(set(languages))

    language_results = yield lang_names_db_thread.submit(get_language_names, locale, db_path)

    if language_results is None:
        logging.error('Failed to locate language name DB: %s', db_path)
        return {}

    converted_languages, duplicated_languages = {}, {}

    for language in languages:
        if language in iso639_codes and iso639_codes[language] in languages:
            duplicated_languages[iso639_codes[language]] = language
            duplicated_languages[language] = iso639_codes[language]

        converted_languages[to_alpha2_code(language)] = language

    output = {}

    if languages:
        for language_result in language_results:
            if language_result[2] in converted_languages:
                language, language_name = language_result[2], language_result[3]
                output[converted_languages[language]] = language_name

                if language in duplicated_languages:
                    output[language] = language_name
                    output[duplicated_languages[language]] = language_name
    else:
        for language_result in language_results:
            output[language_result[2]] = language_result[3]

    return output


class ListLanguageNamesHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        locale_arg = self.get_argument('locale', default=None)
        languages_arg = self.get_argument('languages', default=None)

        if not self.lang_names:
            self.send_response({})
            return

        if locale_arg:
            if languages_arg:
                result = yield get_localized_languages(locale_arg, self.lang_names, languages=languages_arg.split(' '))
            else:
                result = yield get_localized_languages(locale_arg, self.lang_names)

            self.send_response(result)
            return

        if 'Accept-Language' in self.request.headers:
            locales = [locale.split(';')[0].strip() for locale in self.request.headers['Accept-Language'].split(',')]

            for locale in locales:
                result = yield get_localized_languages(locale, self.lang_names)

                if result:
                    self.send_response(result)
                    return

        result = yield get_localized_languages('en', self.lang_names)
        self.send_response(result)



from datetime import timedelta
import re

from tornado import gen

try:
    import fasttext
except ImportError:
    fasttext = None
try:
    import cld2full as cld2  # type: ignore
except ImportError:
    cld2 = None

from apertium_apy.handlers.base import BaseHandler
from apertium_apy.utils import get_coverages, to_alpha3_code


def fasttext_strip_prefix(s):
    """Remove the initial __label__ prefix"""
    return s[9:]


fasttext_max_input = 2048

# there's no [:punct:] class in re module, include the most common here:
fasttext_punct_class = re.compile(r'([`~!@#$%^&*()_=+\[\]{}\\\|;:\"\'<>.,/?—–-]+)')


def fasttext_clean(s):
    """Should clean as ft-train/clean does; also keep input to one line."""
    return re.sub(fasttext_punct_class, r' \1 ',
                  s.lower().replace('\n', ' '))


def fasttext_identify(model, text):
    cleaned = fasttext_clean(text[:fasttext_max_input])
    # Grab a bunch of results since currently the model might predict stuff outside possible_langs – it's still fast:
    results = model.predict(cleaned, k=200, threshold=0.001)
    if results[0]:
        possible_langs = zip(map(fasttext_strip_prefix, results[0]),
                             results[1])
        return {to_alpha3_code(possible_lang[0]): possible_lang[1]
                for possible_lang in possible_langs}
    else:
        return {'nob': 1.0}  # TODO: better default


def cld_identify(text):
    cld_results = cld2.detect(text)
    if cld_results[0]:
        possible_langs = filter(lambda x: x[1] != 'un', cld_results[2])
        return {to_alpha3_code(possible_lang[1]): possible_lang[2]
                for possible_lang in possible_langs}
    else:
        return {'nob': 1.0}  # TODO: better default


class IdentifyLangHandler(BaseHandler):
    fasttext = None

    @gen.coroutine
    def get(self):
        text = self.get_argument('q')
        if not text:
            return self.send_error(400, explanation='Missing q argument')

        if self.fasttext is not None:
            self.send_response(fasttext_identify(self.fasttext, text))
        elif cld2:
            self.send_response(cld_identify(text))
        else:
            try:
                coverages = yield gen.with_timeout(
                    timedelta(seconds=self.timeout),
                    get_coverages(text, self.analyzers, penalize=True),
                )
                self.send_response(coverages)

            except gen.TimeoutError:
                self.send_error(408, explanation='Request timed out')



import logging
import sqlite3
import threading
from collections import defaultdict
from contextlib import closing
from datetime import datetime

if False:
    from typing import Dict  # noqa: F401


class MissingDb(object):
    def __init__(self, db_path, wordmemlimit):
        self.lock = threading.RLock()
        self.conn = None
        self.db_path = db_path
        self.words = defaultdict(lambda: defaultdict(lambda: 0))  # type: Dict[str, Dict[str, int]]
        self.wordcount = 0
        self.wordmemlimit = wordmemlimit

    def note_unknown(self, token, pair):
        self.words[pair][token] += 1
        self.wordcount += 1
        # so if wordmemlimit is 0, we commit on each word
        if self.wordcount > self.wordmemlimit:
            self.commit()
            self.words.clear()
            self.wordcount = 0

    def commit(self):
        time_before = datetime.now()
        with self.lock:
            if not self.conn:
                self.conn = sqlite3.connect(self.db_path)

            with closing(self.conn.cursor()) as c:
                c.execute('PRAGMA synchronous = NORMAL')
                c.execute('CREATE TABLE IF NOT EXISTS missingFreqs (pair TEXT, token TEXT, frequency INTEGER, UNIQUE(pair, token))')
                c.executemany(
                    """INSERT OR REPLACE INTO missingFreqs VALUES (
                        :pair,
                        :token,
                        COALESCE((SELECT frequency FROM missingFreqs WHERE pair=:pair AND token=:token), 0) + :amount
                    )""",
                    ({'pair': pair,
                      'token': token,
                      'amount': self.words[pair][token]}
                        for pair in self.words
                        for token in self.words[pair]))
            self.conn.commit()
        ms = timedelta_to_milliseconds(datetime.now() - time_before)
        logging.info('\tSaving %s unknown words to the DB (%s ms)', self.wordcount, ms)

    def close_db(self):
        if not self.conn:
            logging.warning('no connection on closeDb')
            return
        logging.warning('closing connection')
        self.conn.commit()
        self.conn.close()
        self.conn = None


def timedelta_to_milliseconds(td):
    return td.days * 86400000 + td.seconds * 1000 + int(td.microseconds / 1000)



import tornado.gen

from apertium_apy.handlers.base import BaseHandler
from apertium_apy.utils import to_alpha2_code


class ListHandler(BaseHandler):
    @tornado.gen.coroutine
    def get(self):
        query = self.get_argument('q', default='pairs')

        if query == 'pairs':
            src = self.get_argument('src', default=None)
            response_data = []
            if src:
                pairs = [(src, trg) for trg in self.paths[src]]
            else:
                pairs = [(p[0], p[1]) for par in self.pairs for p in [par.split('-')]]
            for (l1, l2) in pairs:
                response_data.append({'sourceLanguage': l1, 'targetLanguage': l2})
                if self.get_arguments('include_deprecated_codes'):
                    response_data.append({'sourceLanguage': to_alpha2_code(l1), 'targetLanguage': to_alpha2_code(l2)})
            self.send_response({'responseData': response_data, 'responseDetails': None, 'responseStatus': 200})
        elif query == 'analyzers' or query == 'analysers':
            self.send_response({pair: modename for (pair, (path, modename)) in self.analyzers.items()})
        elif query == 'generators':
            self.send_response({pair: modename for (pair, (path, modename)) in self.generators.items()})
        elif query == 'taggers' or query == 'disambiguators':
            self.send_response({pair: modename for (pair, (path, modename)) in self.taggers.items()})
        elif query == 'spellers':
            self.send_response({lang_src: modename for (lang_src, (path, modename)) in self.spellers.items()})
        elif query == 'guessers':
            self.send_response({lang_src: modename for (lang_src, (path, modename)) in self.guessers.items()})

        else:
            self.send_error(400, explanation='Expecting q argument to be one of analysers, generators, guessers, spellers, disambiguators, or pairs')



from apertium_apy.handlers.analyze import AnalyzeHandler  # noqa: F401
from apertium_apy.handlers.base import BaseHandler  # noqa: F401
from apertium_apy.handlers.coverage import CoverageHandler  # noqa: F401
from apertium_apy.handlers.generate import GenerateHandler  # noqa: F401
from apertium_apy.handlers.guesser import GuesserHandler  # noqa: F401
from apertium_apy.handlers.identify_lang import IdentifyLangHandler  # noqa: F401
from apertium_apy.handlers.list_language_names import ListLanguageNamesHandler  # noqa: F401
from apertium_apy.handlers.list_modes import ListHandler  # noqa: F401
from apertium_apy.handlers.per_word import PerWordHandler  # noqa: F401
from apertium_apy.handlers.pipe_debug import PipeDebugHandler  # noqa: F401
from apertium_apy.handlers.speller import SpellerHandler  # noqa: F401
from apertium_apy.handlers.stats import StatsHandler  # noqa: F401
from apertium_apy.handlers.suggestion import SuggestionHandler  # noqa: F401
from apertium_apy.handlers.translate import TranslateHandler  # noqa: F401
from apertium_apy.handlers.translate import PairPrefsHandler  # noqa: F401
from apertium_apy.handlers.translate_chain import TranslateChainHandler  # noqa: F401
from apertium_apy.handlers.translate_doc import TranslateDocHandler  # noqa: F401
from apertium_apy.handlers.translate_raw import TranslateRawHandler  # noqa: F401
from apertium_apy.handlers.translate_webpage import TranslateWebpageHandler  # noqa: F401



from tornado import gen
from tornado.escape import utf8

from apertium_apy.handlers.translate import TranslateHandler


class TranslateRawHandler(TranslateHandler):
    """Assumes the pipeline itself outputs as JSON"""
    def send_response(self, data):
        translated_text = data.get('responseData', {}).get('translatedText', {})
        if translated_text == {}:
            super().send_response(data)
        else:
            self.log_vmsize()
            translated_text = data.get('responseData', {}).get('translatedText', {})
            self.set_header('Content-Type', 'application/json; charset=UTF-8')
            self._write_buffer.append(utf8(translated_text))
            self.finish()

    @gen.coroutine
    def get(self):
        pair = self.get_pair_or_error(self.get_argument('langpair'),
                                      len(self.get_argument('q', strip=False)))
        if pair is not None:
            pipeline = self.get_pipeline(pair)
            deformat = self.get_argument('deformat', default='True') != 'False'
            yield self.translate_and_respond(pair,
                                             pipeline,
                                             self.get_argument('q', strip=False),
                                             self.get_argument('markUnknown', default='yes'),
                                             nosplit=False,
                                             deformat=deformat,
                                             reformat=False)



import re

from tornado import gen

from apertium_apy.handlers.base import BaseHandler
from apertium_apy.utils import to_alpha3_code, remove_dot_from_deformat
from apertium_apy.utils.translation import translate_simple


class AnalyzeHandler(BaseHandler):
    @staticmethod
    def postproc_text(in_text, result):
        lexical_units = remove_dot_from_deformat(in_text, re.findall(r'\^([^\$]*)\$([^\^]*)', result))  # TODO: replace with streamparser
        return [(lu[0], lu[0].split('/')[0] + lu[1])
                for lu
                in lexical_units]

    @gen.coroutine
    def get(self):
        in_text = self.get_argument('q')
        in_mode = to_alpha3_code(self.get_argument('lang'))
        if '-' in in_mode:
            l1, l2 = map(to_alpha3_code, in_mode.split('-', 1))
            in_mode = '%s-%s' % (l1, l2)
        in_mode = self.find_fallback_mode(in_mode, self.analyzers)
        if in_mode in self.analyzers:
            [path, mode] = self.analyzers[in_mode]
            formatting = 'txt'
            commands = [['apertium', '-d', path, '-f', formatting, mode]]
            result = yield translate_simple(in_text, commands)
            self.send_response(self.postproc_text(in_text, result))
        else:
            self.send_error(400, explanation='That mode is not installed')



import html
import logging
import os
import re
import time
from hashlib import sha1
from urllib.parse import urlparse, urlunsplit

from tornado import gen
from tornado import httpclient
from typing import Optional  # noqa: F401


try:
    import chardet
except ImportError:             # type should be Optional[Module] but there's no module in mypy? TODO
    chardet = None              # type: ignore


from apertium_apy.utils import translation
from apertium_apy.handlers.translate import TranslateHandler


class TranslateWebpageHandler(TranslateHandler):
    def url_repl(self, base, attr, quote, aurl):
        a = urlparse(aurl)
        if a.netloc == '':
            newurl = urlunsplit((base.scheme,
                                 base.netloc,
                                 a.path,
                                 a.query,
                                 a.fragment))
        else:
            newurl = aurl
        return ' {a}={q}{u}{q}'.format(a=attr, u=newurl, q=quote)

    def unescape(self, page):
        # First workaround old bug that exists in a lot of
        # Windows-based web pages, see
        # http://stackoverflow.com/a/1398921/69663 :
        page = page.replace('&#150;', '&#8211;')
        # Unescape all other entities the regular way:
        return html.unescape(page)

    def clean_html(self, page, urlbase):
        page = self.unescape(page)
        if urlbase.netloc in ['www.avvir.no', 'avvir.no']:
            page = re.sub(r'([a-zæøåášžđŋ])=([a-zæøåášžđŋ])',
                          '\\1\\2',
                          page)
        page = page.replace('\u00ad', '')  # soft hyphen
        return page

    def html_to_text(self, page, url):
        encoding = 'utf-8'
        if chardet:
            encoding = chardet.detect(page).get('encoding', 'utf-8') or encoding
        base = urlparse(url)
        text = self.clean_html(page.decode(encoding), base)  # type: str
        return re.sub(r' (href|src)=([\'"])(..*?)\2',
                      lambda m: self.url_repl(base, m.group(1), m.group(2), m.group(3)),
                      text)

    def set_cached(self, pair, url, translated, origtext):
        """Cache translated text for a pair and url to memory, and disk.
        Also caches origtext to disk; see cache_path."""
        if pair not in self.url_cache:
            self.url_cache[pair] = {}
        elif len(self.url_cache[pair]) > self.max_inmemory_url_cache:
            self.url_cache[pair] = {}
        ts = time.strftime('%a, %d %b %Y %H:%M:%S GMT', time.gmtime(time.time()))
        self.url_cache[pair][url] = (ts, translated)
        if self.url_cache_path is None:
            logging.info('No --url-cache-path, not storing cached url to disk')
            return
        dirname, basename = self.cache_path(self.url_cache_path, pair, url)
        os.makedirs(dirname, exist_ok=True)
        statvfs = os.statvfs(dirname)
        if (statvfs.f_frsize * statvfs.f_bavail) < self.min_free_space_disk_url_cache:
            logging.warn('Disk of --url-cache-path has < {} free, not storing cached url to disk'.format(
                self.min_free_space_disk_url_cache))
            return
        # Note: If we make this a @gen.coroutine, we will need to lock
        # the file to avoid concurrent same-url requests clobbering:
        path = os.path.join(dirname, basename)
        with open(path, 'w') as f:
            f.write(ts)
            f.write('\n')
            f.write(translated)
        origpath = os.path.join(dirname, pair[0])
        with open(origpath, 'w') as f:
            f.write(origtext)

    def cache_path(self, url_cache_path, pair, url):
        """Give the directory for where to cache the translation of this url,
        and the file name to use for this pair."""
        hsh = sha1(url.encode('utf-8')).hexdigest()
        dirname = os.path.join(url_cache_path,
                               # split it to avoid too many files in one dir:
                               hsh[:1], hsh[1:2], hsh[2:])
        return (dirname, '{}-{}'.format(*pair))

    def get_cached(self, pair, url):
        if not self.url_cache_path:
            return None
        if pair not in self.url_cache:
            self.url_cache[pair] = {}
        if url in self.url_cache[pair]:
            logging.info('Got cache from memory')
            return self.url_cache[pair][url]
        dirname, basename = self.cache_path(self.url_cache_path, pair, url)
        path = os.path.join(dirname, basename)
        if os.path.exists(path):
            logging.info('Got cache on disk, we want to retranslate in background …')
            with open(path, 'r') as f:
                return (f.readline().strip(), f.read())

    def retranslate_cache(self, pair, url, cached):
        """If we've got something from the cache, and it isn't in memory, then
        it was from disk. We want to retranslate anything we found on
        disk, since it's probably using older versions of the language
        pair.
        """
        mem_cached = self.url_cache.get(pair, {}).get(url)
        if mem_cached is None and cached is not None and self.url_cache_path is not None:
            dirname, _ = self.cache_path(self.url_cache_path, pair, url)
            origpath = os.path.join(dirname, pair[0])
            if os.path.exists(origpath):
                return open(origpath, 'r').read()

    @gen.coroutine
    def get(self):
        pair = self.get_pair_or_error(self.get_argument('langpair'),
                                      # Don't yet know the size of the text, and don't want to fetch it unnecessarily:
                                      -1)
        if pair is None:
            return
        self.note_pair_usage(pair)
        prefs = self.get_argument('prefs', default='')
        mode_path = self.pairs['%s-%s' % pair]
        url = self.get_argument('url')
        if not url.startswith('http'):
            url = 'http://' + url
        got304 = False
        cached = self.get_cached(pair, url)
        if prefs:
            logging.warn("Web translation with prefs doesn't work with caching yet")
        request = httpclient.HTTPRequest(url=url,
                                         # TODO: tweak timeouts:
                                         connect_timeout=20.0,
                                         request_timeout=20.0)

        try:
            response = yield httpclient.AsyncHTTPClient().fetch(request)
        except Exception as e:
            logging.info('%s exception has occurred', e)
            self.send_error(404, explanation='{} on fetching url: {}'.format('Error 404', e))
            return
        try:
            response = yield httpclient.AsyncHTTPClient().fetch(request, raise_error=True)
        except httpclient.HTTPError as e:
            if e.code == 304:
                got304 = True
                logging.info('304, can use cache')
            else:
                logging.error(e)
                self.send_error(503, explanation='{} on fetching url: {}'.format(response.code, response.error))
                return
        if got304 and cached is not None:
            translation_catpipeline = translation.CatPipeline  # type: ignore
            translated = yield translation_catpipeline().translate(cached[1])
        else:
            if response.body is None:
                self.send_error(503, explanation='got an empty file on fetching url: {}'.format(url))
                return
            page = response.body  # type: bytes
            try:
                to_translate = self.html_to_text(page, url)
            except UnicodeDecodeError as e:
                logging.info("/translatePage '{}' gave UnicodeDecodeError {}".format(url, e))
                self.send_error(503, explanation="Couldn't decode (or detect charset/encoding of) {}".format(url))
                return
            before = self.log_before_translation()
            translated = yield translation.translate_html_mark_headings(to_translate, mode_path, prefs)
            self.log_after_translation(before, len(to_translate))
            self.set_cached(pair, url, translated, to_translate)
        self.send_response({
            'responseData': {
                'translatedText': self.maybe_strip_marks(self.mark_unknown, pair, translated),
            },
            'responseDetails': None,
            'responseStatus': 200,
        })
        retranslate = self.retranslate_cache(pair, url, cached)
        if got304 and retranslate is not None:
            logging.info('Retranslating {}'.format(url))
            translated = yield translation.translate_html_mark_headings(retranslate, mode_path, prefs)
            logging.info('Done retranslating {}'.format(url))
            self.set_cached(pair, url, translated, retranslate)



from tornado import gen

from apertium_apy.handlers.base import BaseHandler
from apertium_apy.utils import to_alpha3_code
from apertium_apy.utils.translation import parse_mode_file, translate_pipeline


class PipeDebugHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        to_translate = self.get_argument('q')

        try:
            l1, l2 = map(to_alpha3_code, self.get_argument('langpair').split('|'))
        except ValueError:
            self.send_error(400, explanation='That pair is invalid, use e.g. eng|spa')
            return

        mode_path = self.pairs['%s-%s' % (l1, l2)]
        try:
            _, commands = parse_mode_file(mode_path)
        except Exception:
            self.send_error(500)
            return

        res = yield translate_pipeline(to_translate, commands)
        if self.get_status() != 200:
            self.send_error(self.get_status())
            return

        output, pipeline = res

        self.send_response({
            'responseData': {'output': output, 'pipeline': pipeline},
            'responseDetails': None,
            'responseStatus': 200,
        })



import heapq
import logging
import re
import time
from datetime import datetime

from tornado import gen
import tornado.iostream
import asyncio

from apertium_apy import missing_freqs_db  # noqa: F401
from apertium_apy.handlers.base import BaseHandler
from apertium_apy.keys import ApiKeys
from apertium_apy.utils import to_alpha3_code, scale_mt_log
from apertium_apy.utils.translation import parse_mode_file, make_pipeline
# Typing imports that flake8 doesn't understand:
from apertium_apy.utils.translation import FlushingPipeline, SimplePipeline  # noqa: F401
from typing import Union        # noqa: F401


class TranslationInfo:
    def __init__(self, handler):
        self.langpair = handler.get_argument('langpair')
        self.key = handler.get_argument('key', default='null')
        self.ip = handler.request.headers.get('X-Real-IP', handler.request.remote_ip)
        self.referer = handler.request.headers.get('Referer', 'null')


class TranslateHandler(BaseHandler):
    unknown_mark_re = re.compile(r'[*]([^.,;:\t\* ]+)')
    api_keys = None

    def __init__(self, application, request, **kwargs):
        super().__init__(application, request, **kwargs)

    @property
    def mark_unknown(self):
        return self.get_argument('markUnknown', default='yes').lower() in ['yes', 'true', '1']

    def note_pair_usage(self, pair):
        self.stats.usecount[pair] = 1 + self.stats.usecount.get(pair, 0)

    def maybe_strip_marks(self, mark_unknown, pair, translated):
        self.note_unknown_tokens('%s-%s' % pair, translated)
        if mark_unknown:
            return translated
        else:
            return re.sub(self.unknown_mark_re, r'\1', translated)

    def note_unknown_tokens(self, pair, text):
        global missing_freqs_db
        if missing_freqs_db is not None:
            for token in re.findall(self.unknown_mark_re, text):
                missing_freqs_db.note_unknown(token, pair)

    def cleanable(self, i, pair, pipe):
        if pipe.stuck:
            logging.info('A pipe for pair %s-%s seems stuck, scheduling restart',
                         pair[0], pair[1])
            return True
        if pipe.use_count > self.restart_pipe_after:
            # Not affected by min_pipes_per_pair
            logging.info('A pipe for pair %s-%s has handled %d requests, scheduling restart',
                         pair[0], pair[1], self.restart_pipe_after)
            return True
        elif (i >= self.min_pipes_per_pair and
                self.max_idle_secs != 0 and
                time.time() - pipe.last_usage > self.max_idle_secs):
            logging.info("A pipe for pair %s-%s hasn't been used in %d secs, scheduling shutdown",
                         pair[0], pair[1], self.max_idle_secs)
            return True
        else:
            return False

    def clean_pairs(self):
        for pair in self.pipelines:
            pipes = self.pipelines[pair]
            to_clean = set(p for i, p in enumerate(pipes)
                           if self.cleanable(i, pair, p))
            self.pipelines_holding += to_clean
            pipes[:] = [p for p in pipes if p not in to_clean]
            heapq.heapify(pipes)
        # The holding area lets us restart pipes after n usages next
        # time round, since with lots of traffic an active pipe may
        # never reach 0 users
        self.pipelines_holding[:] = [p for p in self.pipelines_holding
                                     if p.users > 0]
        if self.pipelines_holding:
            logging.info('%d pipelines still scheduled for shutdown', len(self.pipelines_holding))

    def get_pipe_cmds(self, l1, l2):
        if (l1, l2) not in self.pipeline_cmds:
            mode_path = self.pairs['%s-%s' % (l1, l2)]
            self.pipeline_cmds[(l1, l2)] = parse_mode_file(mode_path)
        return self.pipeline_cmds[(l1, l2)]

    def should_start_pipe(self, l1, l2):
        pipes = self.pipelines.get((l1, l2), [])
        if pipes == []:
            logging.info('%s-%s not in pipelines of this process',
                         l1, l2)
            return True
        else:
            min_p = pipes[0]
            if len(pipes) < self.max_pipes_per_pair and min_p.users > self.max_users_per_pipe:
                logging.info('%s-%s has ≥%d users per pipe but only %d pipes',
                             l1, l2, min_p.users, len(pipes))
                return True
            else:
                return False

    def get_pipeline(self, pair):
        (l1, l2) = pair
        if self.should_start_pipe(l1, l2):
            logging.info('Starting up a new pipeline for %s-%s …', l1, l2)
            if pair not in self.pipelines:
                self.pipelines[pair] = []
            p = make_pipeline(self.get_pipe_cmds(l1, l2), self.timeout)
            heapq.heappush(self.pipelines[pair], p)
        return self.pipelines[pair][0]

    def log_before_translation(self):
        return datetime.now()

    def log_after_translation(self, before, length):
        after = datetime.now()
        if self.scale_mt_logs:
            t_info = TranslationInfo(self)
            key = self.get_api_key(t_info.key)
            scale_mt_log(self.get_status(), after - before, t_info, key, length)

        if self.get_status() == 200:
            timings = self.stats.timing
            oldest = timings[0][0] if timings else datetime.now()
            if datetime.now() - oldest > self.stat_period_max_age:
                self.stats.timing.pop(0)
            self.stats.timing.append(
                (before, after, length))

    def get_pair_or_error(self, langpair, text_length):
        try:
            l1, l2 = map(to_alpha3_code, langpair.split('|'))
            in_mode = '%s-%s' % (l1, l2)
        except ValueError:
            self.send_error(400, explanation='That pair is invalid, use e.g. eng|spa')
            self.log_after_translation(self.log_before_translation(), text_length)
            return None
        in_mode = self.find_fallback_mode(in_mode, self.pairs)
        if in_mode not in self.pairs:
            self.send_error(400, explanation='That pair is not installed')
            self.log_after_translation(self.log_before_translation(), text_length)
            return None
        else:
            return tuple(in_mode.split('-'))

    def get_format(self):
        dereformat = self.get_argument('format', default=None)
        deformat = ''
        reformat = ''
        if dereformat:
            deformat = 'apertium-des' + dereformat
            reformat = 'apertium-re' + dereformat
        else:
            deformat = self.get_argument('deformat', default='html')
            if 'apertium-des' not in deformat:
                deformat = 'apertium-des' + deformat
            reformat = self.get_argument('reformat', default='html-noent')
            if 'apertium-re' not in reformat:
                reformat = 'apertium-re' + reformat

        return deformat, reformat

    @gen.coroutine
    def translate_and_respond(self, pair, pipeline, to_translate, mark_unknown, nosplit=False, deformat=True, reformat=True, prefs=''):
        mark_unknown = mark_unknown in ['yes', 'true', '1']
        self.note_pair_usage(pair)
        before = self.log_before_translation()
        try:
            translated = yield pipeline.translate(to_translate, nosplit, deformat, reformat, prefs)
            self.log_after_translation(before, len(to_translate))
            self.send_response({
                'responseData': {
                    'translatedText': self.maybe_strip_marks(mark_unknown, pair, translated),
                },
                'responseDetails': None,
                'responseStatus': 200,
            })
        except asyncio.TimeoutError as e:
            logging.warning('Translation error in pair %s-%s: %s', pair[0], pair[1], e)
            pipeline.stuck = True
            self.send_error(503, explanation='internal error')
        except tornado.iostream.StreamClosedError as e:
            logging.warning('Translation error in pair %s-%s: %s', pair[0], pair[1], e)
            pipeline.stuck = True
            self.send_error(503, explanation='internal error')
        self.clean_pairs()

    @gen.coroutine
    def get(self):
        pair = self.get_pair_or_error(self.get_argument('langpair'),
                                      len(self.get_argument('q')))
        if pair is not None:
            pipeline = self.get_pipeline(pair)  # type: Union[FlushingPipeline, SimplePipeline]
            deformat, reformat = self.get_format()
            yield self.translate_and_respond(pair,
                                             pipeline,
                                             self.get_argument('q'),
                                             self.get_argument('markUnknown', default='yes'),
                                             nosplit=False,
                                             deformat=deformat,
                                             reformat=reformat,
                                             prefs=self.get_argument('prefs', default=''),
                                             )

    @classmethod
    def get_api_key(cls, key):
        if not cls.api_keys:
            cls.api_keys = ApiKeys(cls.api_keys_conf)

        return cls.api_keys.get_key(key)


class PairPrefsHandler(BaseHandler):
    @gen.coroutine
    def get(self):
        self.send_response(self.pairprefs)



import logging

from tornado import gen

try:
    import requests
except ImportError:
    requests = None  # type: ignore

from apertium_apy import BYPASS_TOKEN, RECAPTCHA_VERIFICATION_URL
from apertium_apy.handlers.base import BaseHandler
from apertium_apy.utils.wiki import wiki_get_token, wiki_get_page, wiki_add_text, wiki_edit_page

if False:
    from typing import Optional  # noqa: F401


def add_suggestion(s, suggest_url, edit_token, data):
    content = wiki_get_page(s, suggest_url)
    content = wiki_add_text(content, data)
    edit_result = wiki_edit_page(s, suggest_url, content, edit_token)

    try:
        if edit_result['edit']['result'] == 'Success':
            logging.info('Update of page %s', suggest_url)
            return True
        else:
            logging.error('Update of page %s failed: %s', suggest_url, edit_result)
            return False
    except KeyError:
        return False


class SuggestionHandler(BaseHandler):
    wiki_session = None  # type: Optional[requests.Session]
    wiki_edit_token = None
    SUGGEST_URL = None
    recaptcha_secret = None
    auth_token = None

    @gen.coroutine
    def get(self):
        self.send_error(405, explanation='GET request not supported')

    @gen.coroutine
    def post(self):
        context = self.get_argument('context', None)
        word = self.get_argument('word', None)
        new_word = self.get_argument('newWord', None)
        langpair = self.get_argument('langpair', None)
        recap = self.get_argument('g-recaptcha-response', None)

        if not new_word:
            self.send_error(400, explanation='A suggestion is required')
            return

        if not recap:
            self.send_error(400, explanation='The ReCAPTCHA is required')
            return

        if not all([context, word, langpair, new_word, recap]):
            self.send_error(400, explanation='All arguments were not provided')
            return

        logging.info('Suggestion (%s): Context is %s \n Word: %s ; New Word: %s ', langpair, context, word, new_word)
        logging.info('Now verifying ReCAPTCHA.')

        if not self.recaptcha_secret:
            logging.error('No ReCAPTCHA secret provided!')
            self.send_error(400, explanation='Server not configured correctly for suggestions')
            return

        if recap == BYPASS_TOKEN:
            logging.info('Adding data to wiki with bypass token')
        else:
            # for nginx or when behind a proxy
            x_real_ip = self.request.headers.get('X-Real-IP')
            user_ip = x_real_ip or self.request.remote_ip
            payload = {
                'secret': self.recaptcha_secret,
                'response': recap,
                'remoteip': user_ip,
            }
            recap_request = self.wiki_session.post(RECAPTCHA_VERIFICATION_URL, data=payload)
            if recap_request.json()['success']:
                logging.info('ReCAPTCHA verified, adding data to wiki')
            else:
                logging.info('ReCAPTCHA verification failed, stopping')
                self.send_error(400, explanation='ReCAPTCHA verification failed')
                return

        data = {
            'context': context, 'langpair': langpair,
            'word': word, 'newWord': new_word,
        }
        result = add_suggestion(self.wiki_session,
                                self.SUGGEST_URL, self.wiki_edit_token,
                                data)

        if result:
            self.send_response({
                'responseData': {
                    'status': 'Success',
                },
                'responseDetails': None,
                'responseStatus': 200,
            })
        else:
            logging.info('Page update failed, trying to get new edit token')
            self.wiki_edit_token = wiki_get_token(
                SuggestionHandler.wiki_session, 'edit', 'info|revisions')
            logging.info('Obtained new edit token. Trying page update again.')
            result = add_suggestion(self.wiki_session,
                                    self.SUGGEST_URL, self.wiki_edit_token,
                                    data)
            if result:
                self.send_response({
                    'responseData': {
                        'status': 'Success',
                    },
                    'responseDetails': None,
                    'responseStatus': 200,
                })
            else:
                self.send_error(400, explanation='Page update failed')



#!/usr/bin/env python3
# coding=utf-8
# -*- indent-tabs-mode: nil -*-
# -*- encoding: utf-8 -*-

"""
Copyright (C) 2016 Kevin Brubeck Unhammer
based on https://gist.github.com/Spindel/1d07533ef94a4589d348 / watchdogged.py
Copyright (C) 2015 D.S. Ljungmark, Modio AB
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import socket
import os

# All singletons are prefixed "the"
the_log = logging.getLogger(__name__)


def watchdog_period():
    """Return the time (in seconds) that we need to ping within."""
    val = os.environ.get('WATCHDOG_USEC', None)
    if not val:
        return None
    return int(val) / 1000000


def notify_socket(clean_environment=True):
    """Return a tuple of address, socket for future use.
    clean_environment removes the variables from env to prevent children
    from inheriting it and doing something wrong.
    """
    _empty = None, None
    address = os.environ.get('NOTIFY_SOCKET', None)
    if clean_environment:
        address = os.environ.pop('NOTIFY_SOCKET', None)

    if not address:
        return _empty

    if len(address) == 1:
        return _empty

    if address[0] not in ('@', '/'):
        return _empty

    if address[0] == '@':
        address = '\0' + address[1:]

    # SOCK_CLOEXEC was added in Python 3.2 and requires Linux >= 2.6.27.
    # It means "close this socket after fork/exec()
    try:
        sock = socket.socket(socket.AF_UNIX,
                             socket.SOCK_DGRAM | socket.SOCK_CLOEXEC)
    except AttributeError:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)

    return address, sock


class Watchdog(object):

    def __init__(self, period, address, sock):
        # "The daemon should then issue sd_notify("WATCHDOG=1") calls every half of that interval."
        self.period = float(period) / 2.0
        self.address = address
        self.sock = sock

    def __del__(self):
        self.systemd_stop()

    def sd_message(self, message):
        """Send a message to the systemd bus/socket.
        message is expected to be bytes.
        """
        if not (self.address and self.sock and message):
            the_log.info("Couldn't message! {} {} {}".format(self.address, self.sock, message))
            return False
        assert isinstance(message, bytes)

        try:
            retval = self.sock.sendto(message, self.address)
        except socket.error:
            return False
        return (retval > 0)

    def watchdog_ping(self):
        """Helper function to send a watchdog ping."""
        return self.sd_message(b'WATCHDOG=1')

    def systemd_ready(self):
        """Helper function to send a ready signal."""
        return self.sd_message(b'READY=1')

    def systemd_stop(self):
        """Helper function to signal service stopping."""
        return self.sd_message(b'STOPPING=1')


def setup_watchdog():
    # Get our settings from the environment
    notify = notify_socket()
    period = watchdog_period()
    # Validate some in-data
    if not notify[0]:
        the_log.info('No notification socket, not launched via systemd?')
        return None
    if not period:
        the_log.warning('Found systemd notification socket, but no watchdog period set in the unit file!')
        return None
    wd = Watchdog(period, *notify)
    return wd
